      ! Fortran Version = 3
      subroutine check_version(version, warning) 
              implicit none
              integer :: version, warning
cf2py         intent(in) :: version
cf2py         intent(out) :: warning
              if (version .NE. 3) then
                warning = 1
              else
                warning = 0
              end if     
       end subroutine
 
       subroutine calculate_g2(numbers, rs, g_number, g_eta, cutoff, 
     *                                                home, n, ridge)
               
              implicit none
              integer, dimension(n) :: numbers
              integer, dimension(1) :: g_number
              double precision, dimension(n, 3) :: rs
              double precision, dimension(3) :: home
              integer :: n
              double precision ::  g_eta, cutoff
              double precision :: ridge
cf2py         intent(in) :: numbers, rs, g_number
cf2py         intent(in) :: g_eta, cutoff, home
cf2py         intent(hide) :: n
cf2py         intent(out) :: ridge
              integer :: j, match, xyz
              double precision, dimension(3) :: Rij_
              double precision :: Rij, term

              ridge = 0.0d0
              do j = 1, n
                  match = compare(numbers(j), g_number(1))
                  if (match == 1) then
                    do xyz = 1, 3
                      Rij_(xyz) = rs(j, xyz) - home(xyz)
                    end do
                    Rij = sqrt(dot_product(Rij_, Rij_))
                    term = exp(-g_eta*(Rij**2))
                    term = term * cutoff_fxn(Rij, cutoff)
                    ridge = ridge + term
                  end if
                end do

      CONTAINS

      function compare(try, val) result(match)
C     Returns 1 if try is the same set as val, 0 if
C     not.
              implicit none
              integer, intent(in) :: try, val
              integer :: match
              if (try == val) then
                      match = 1
              else
                      match = 0
              end if
      end function compare

      function cutoff_fxn(r, cutoff)
              double precision :: r, cutoff, cutoff_fxn, pi
              if (r > cutoff) then
                      cutoff_fxn = 0.0d0
              else
                      pi = 4.0d0 * datan(1.0d0)
                      cutoff_fxn = 0.5d0 * (cos(pi*r/cutoff) + 1.0d0)
              end if

      end function
      
      end subroutine calculate_g2
       
      subroutine calculate_g4(numbers, rs, g_numbers, g_gamma, g_zeta, 
     *                        g_eta, cutoff, home, n, ridge)
               
              implicit none
              integer, dimension(n) :: numbers
              integer, dimension(2) :: g_numbers
              double precision, dimension(n, 3) :: rs
              double precision, dimension(3) :: home
              integer :: n
              double precision :: g_gamma, g_zeta, g_eta, cutoff
              double precision :: ridge
cf2py         intent(in) :: numbers, rs, g_numbers, g_gamma, g_zeta
cf2py         intent(in) :: g_eta, cutoff, home
cf2py         intent(hide) :: n
cf2py         intent(out) :: ridge
              integer :: j, k, match, xyz
              double precision, dimension(3) :: Rij_, Rik_, Rjk_
              double precision :: Rij, Rik, Rjk, costheta, term

              ridge = 0.0d0
              do j = 1, n
                do k = (j + 1), n
                  match = compare(numbers(j), numbers(k), g_numbers(1),
     *                            g_numbers(2))
                  if (match == 1) then
                    do xyz = 1, 3
                      Rij_(xyz) = rs(j, xyz) - home(xyz)
                      Rik_(xyz) = rs(k, xyz) - home(xyz)
                      Rjk_(xyz) = rs(k, xyz) - rs(j, xyz)
                    end do
                    Rij = sqrt(dot_product(Rij_, Rij_))
                    Rik = sqrt(dot_product(Rik_, Rik_))
                    Rjk = sqrt(dot_product(Rjk_, Rjk_))
                    costheta = dot_product(Rij_, Rik_) / Rij / Rik
                    term = (1.0d0 + g_gamma * costheta)**g_zeta
                    term = term*exp(-g_eta*(Rij**2 + Rik**2 + Rjk**2))
                    term = term * cutoff_fxn(Rij, cutoff)
                    term = term * cutoff_fxn(Rik, cutoff)
                    term = term * cutoff_fxn(Rjk, cutoff)
                    ridge = ridge + term
                  end if
                end do
              end do
              ridge = ridge * 2.0d0**(1.0d0 - g_zeta)


      CONTAINS

      function compare(try1, try2, val1, val2) result(match)
C     Returns 1 if (try1, try2) is the same set as (val1, val2), 0 if
C     not.
              implicit none
              integer, intent(in) :: try1, try2, val1, val2
              integer :: match
              integer :: ntry1, ntry2, nval1, nval2
              ! First sort to avoid endless logical loops.
              if (try1 < try2) then
                      ntry1 = try1
                      ntry2 = try2
              else
                      ntry1 = try2
                      ntry2 = try1
              end if
              if (val1 < val2) then
                      nval1 = val1
                      nval2 = val2
              else
                      nval1 = val2
                      nval2 = val1
              end if
              if (ntry1 == nval1 .AND. ntry2 == nval2) then
                      match = 1
              else
                      match = 0
              end if

      end function compare

      function cutoff_fxn(r, cutoff)
              double precision :: r, cutoff, cutoff_fxn, pi
              if (r > cutoff) then
                      cutoff_fxn = 0.0d0
              else
                      pi = 4.0d0 * datan(1.0d0)
                      cutoff_fxn = 0.5d0 * (cos(pi*r/cutoff) + 1.0d0)
              end if

      end function
      
      end subroutine calculate_g4
      
      subroutine process_o(o_input, weight_no_row, weight_no_col, 
     * weight, activation_function, o_output)

            implicit none
            integer :: weight_no_row, weight_no_col
            double precision, dimension (weight_no_row - 1) :: o_input
          double precision, dimension (weight_no_row, weight_no_col) :: 
     *  weight
             character(len=*) :: activation_function  
            double precision, dimension (weight_no_col) :: o_output
!f2py   intent (in) :: o_input    
!f2py   intent (in) :: weight_no_row
!f2py   intent (in) :: weight_no_col
!f2py  intent (in) :: weight
!f2py  intent (in) :: activation_function
!f2py  intent (out) :: o_output
            double precision, dimension (weight_no_row) :: ohat
            double precision, dimension (weight_no_col) :: net
             integer :: i, j

            !o[0] = indexfp
            ohat = 0.0d0
            do i = 1, weight_no_row - 1
                    ohat(i) = o_input(i)
            enddo
            ohat(weight_no_row) = 1.0d0  ! one bias
            net = 0.0d0
            o_output = 0.0d0
            do i = 1, weight_no_col
                do j = 1, weight_no_row
                    net(i) = net(i) + ohat(j) * weight(j,i)
                enddo
                if (activation_function .NE. 'linear' .AND. 
     * activation_function .NE. 'tanh' .AND. 
     * activation_function .NE. 'sigmoid')   then
                print *, "Fortran Error: Unknown activation function!"
                else if (activation_function .EQ. 'linear') then
                o_output(i) = net(i)
                else if (activation_function .EQ. 'tanh') then
                o_output(i) = tanh(net(i))
                else if (activation_function .EQ. 'sigmoid') then    
                o_output(i) = 1./(1. + exp(-net(i)))
                end if
            enddo
    
      end subroutine process_o
      
            subroutine calculate_error(no_atom, nn_values, real_values,
     *            no_values, rmse_per_atom, square_error)

      implicit none
      double precision:: rmse_per_atom, square_error
      integer :: no_atom, no_values
      double precision, dimension (no_values) :: nn_values,
     *        real_values
             
!f2py   intent (in) :: no_atom, no_values, nn_values, real_values
!f2py  intent (out) :: rmse_per_atom, square_error

      integer :: i
      double precision :: square_error_per_image
      square_error = 0.0d0
      do i = 1, no_values
      square_error = square_error +
     * (nn_values(i) - real_values(i))**(2.0d0)
      enddo
      square_error_per_image = square_error / no_values
      rmse_per_atom = sqrt(square_error_per_image) / no_atom
      end subroutine calculate_error
      
         subroutine der_parameters(e_weight, e_scaling, number, numbers,
     *  len_numbers, indices, no_of_elements,
     *  elements_numbers, no_nodes_of_elements,
     * len_no_nodes_of_elements, vector_outputs, len_vector_outputs,
     * vector_W, len_vector_W,activation_function,no_layers_of_elements,
     * no_all_nodes_of_elements, no_weights_of_element,
     * der_weights, der_scaling_intercept, der_scaling_slope)

            implicit none
            double precision :: e_weight, e_scaling
            double precision :: der_scaling_intercept, der_scaling_slope
            integer :: number, len_numbers, no_of_elements
            integer :: len_vector_W, len_vector_outputs
            integer :: len_no_nodes_of_elements, no_weights_of_element
            integer, dimension (no_of_elements) :: elements_numbers
            integer, dimension(no_of_elements)::no_all_nodes_of_elements
            double precision, dimension (len_vector_outputs) ::
     *  vector_outputs
            double precision, dimension (len_vector_W) ::
     *  vector_W
            double precision, dimension (no_weights_of_element) ::
     *  der_weights
     
            integer, dimension (no_of_elements) :: no_layers_of_elements
            integer, dimension (len_numbers) :: numbers, indices
            integer, dimension (len_no_nodes_of_elements) ::
     *          no_nodes_of_elements
             character(len=*) :: activation_function
             
!f2py   intent (in) :: number, numbers,  indices, no_of_elements, elements_numbers, e
!f2py   intent (in) :: total_no_layers_of_elements, no_nodes_of_elements, vector_W
!f2py   intent (in) :: len_no_nodes_of_elements, vector_outputs, len_vector_outputs, no_all_nodes_of_elements
!f2py  intent (in) :: len_vector_W, activation_function, len_numbers, no_layers_of_elements, no_weights_of_element
!f2py  intent (out) :: der_weights, der_scaling_intercept, der_scaling_slope

            integer :: atom, index, symbol, position
            integer :: W_rows, W_cols, count, j
            integer :: i, N, k, index_symbol_no, layer_no
            integer :: len_unravel, ohat_size
            double precision, allocatable:: delta(:,:)
            double precision, allocatable:: temp(:,:)
            double precision, allocatable:: W(:,:), weight_der(:,:)
           double precision, allocatable:: D(:,:), ohat(:, :),unravel(:)
      
      der_weights = 0.0d0
      der_scaling_intercept = 0.0d0
      der_scaling_slope = 0.0d0
      
       do atom = 1, len_numbers
            index = indices(atom) + 1
            symbol = numbers(atom)
            if (symbol == number) then
                   der_scaling_intercept = 
     *            der_scaling_intercept + 2.0d0 * e_scaling
                    index_symbol_no = symbol_no(symbol)
                    N = no_layers_of_elements(index_symbol_no) - 2
                    layer_no =
     *                no_prev_layers_of_all_elements(index_symbol_no)
                    k = N + 1
                    position = 1
                    position = position + no_nodes_of_prev_atoms(index)
                    position = position +
     *                      no_nodes_of_prev_layers(layer_no, k + 1)
                    der_scaling_slope = der_scaling_slope
     *               + 2.0d0 * e_scaling * vector_outputs(position)
                    do while (k .GT. 1)                    
                        if (k == N + 1) then
                            if (allocated(D) .EQV. .True.)
     *                     deallocate(D)
                           allocate(D(1, 1))
                           D = calculate_D(1, index, layer_no, k + 1)
                            if (allocated(delta) .EQV. .True.)
     *                     deallocate(delta)
                            allocate(delta(1, 1))
                            delta = D 
                            !calculating W
                          W_rows = no_nodes_of_elements(layer_no + k)
                            W_cols = no_nodes_of_elements(layer_no +k+1)
                            if (allocated(W) .EQV. .True.)
     *                     deallocate(W)
                           allocate(W(W_rows, W_cols))
                           W = matrix_W(W_rows, W_cols, layer_no, k)
                         ! calculating ohat
                         ohat_size = W_rows + 1
                         if (allocated(ohat) .EQV. .True.)
     *                     deallocate(ohat)
                         allocate(ohat(ohat_size, 1))
                         ohat = 0.0d0
                        do i = 1, ohat_size - 1
                                position = i
                                position = position +
     *                         no_nodes_of_prev_atoms(index)
                                position = position +
     *                         no_nodes_of_prev_layers(layer_no, k)
                                ohat(i, 1) =  vector_outputs(position)
                         enddo
                          ohat(ohat_size, 1) = 1.0d0
                        ! calculating weight_der  
                        if (allocated(weight_der) .EQV. .True.)
     *                   deallocate(weight_der)
                        allocate(weight_der(ohat_size, W_cols))
                          weight_der = e_weight
     *                     *matmul(ohat, transpose(delta))
                          if (allocated(unravel) .EQV. .True.)
     *                     deallocate(unravel)
                            allocate(unravel(ohat_size*W_cols))
                            unravel=
     *                     unravel_matrix(weight_der,ohat_size, W_cols)
                           len_unravel = size(unravel)
                           count = 0
                           do i = 1, k - 1
                                count = count +
     *                         (no_nodes_of_elements(layer_no + i) + 1)*
     *                         no_nodes_of_elements(layer_no + i + 1)
                           enddo
                           do i = 1, len_unravel
                              der_weights(count + i) = 
     *                       der_weights(count + i) + unravel(i)
                          enddo              
                          k = k - 1
                       else
                            !calculating new D
                            if (allocated(D) .EQV. .True.)
     *                     deallocate(D)
                            allocate(D(W_rows, W_rows))
                            D = calculate_D(W_rows, index, layer_no,k+1)
                            ! calculating new delta using new D, old W, and old delta
                            if (allocated(temp) .EQV. .True.)
     *                       deallocate(temp)
                            allocate(temp(W_rows, 1))
                            temp = matmul(W, delta)
                            if (allocated(delta) .EQV. .True.)
     *                     deallocate(delta)
                            allocate(delta(W_cols, 1))
                            delta = matmul(D, temp)
                            !calculating new W
                          W_rows = no_nodes_of_elements(layer_no + k)
                            W_cols = no_nodes_of_elements(layer_no +k+1)
                            if (allocated(W) .EQV. .True.)
     *                     deallocate(W)
                            allocate(W(W_rows, W_cols))
                           W = matrix_W(W_rows, W_cols, layer_no, k)
                          ! calculating new ohat
                          ohat_size = W_rows + 1
                         if (allocated(ohat) .EQV. .True.)
     *                     deallocate(ohat)
                         allocate(ohat(ohat_size, 1))
                         ohat = 0.0d0
                        do i = 1, ohat_size - 1
                                position = i
                                position = position +
     *                         no_nodes_of_prev_atoms(index)
                                position = position +
     *                         no_nodes_of_prev_layers(layer_no, k)
                                ohat(i, 1) =  vector_outputs(position)
                         enddo
                          ohat(ohat_size, 1) = 1.0d0
                        ! calculating weight_der  
                        if (allocated(weight_der) .EQV. .True.)
     *                   deallocate(weight_der)
                        allocate(weight_der(ohat_size, W_cols))
                          weight_der = e_weight
     *                     *matmul(ohat, transpose(delta))
                          if (allocated(unravel) .EQV. .True.)
     *                     deallocate(unravel)
                            allocate(unravel(ohat_size*W_cols))
                            unravel=
     *                     unravel_matrix(weight_der,ohat_size, W_cols)
                           len_unravel = size(unravel)
                           count = 0
                           do i = 1, k - 1
                                count = count +
     *                         (no_nodes_of_elements(layer_no + i) + 1)*
     *                         no_nodes_of_elements(layer_no + i + 1)
                           enddo
                            do i = 1, len_unravel
                              der_weights(count + i) = 
     *                       der_weights(count + i) + unravel(i)
                          enddo
                           k = k - 1
                       endif              
                    enddo
          endif 
          enddo
      
        CONTAINS
        
      function calculate_D(d_size, index, layer_no, layer)
      implicit none
      integer :: d_size, index, layer_no, layer
      double precision, dimension(d_size , d_size):: calculate_D
            
           calculate_D = 0.0d0
           do i = 1, d_size
                 position = i
                 position = position + no_nodes_of_prev_atoms(index)
                 position = position +
     *                      no_nodes_of_prev_layers(layer_no, layer)
                 if (activation_function .EQ. 'linear') then
                       calculate_D(i,i) = 1.0d0
                 else if (activation_function .EQ. 'tanh')  then
                    calculate_D(i,i)  = 1.0d0 - vector_outputs(position)
     *                        * vector_outputs(position)
                 else if (activation_function .EQ. 'sigmoid') then
                  calculate_D(i,i)  = (1.0d0 - vector_outputs(position))
     *                * vector_outputs(position)
                 endif
            enddo
      end function calculate_D
      
      function matrix_W(W_rows , W_cols, layer_no, layer)
      implicit none
      integer :: layer_no, layer, W_rows , W_cols
      double precision, dimension(W_rows , W_cols):: matrix_W
      
           count = 0
           do i = 1, layer - 1
                count = count + no_nodes_of_elements(layer_no + i)*
     *                          no_nodes_of_elements(layer_no + i + 1)
           enddo
           matrix_W = 0.0d0 
           do i = 1, W_rows
               do j = 1, W_cols
                   matrix_W(i,j) =
     *              vector_W(count + W_cols*(i - 1) + j)
               enddo
           enddo       
      end function matrix_W
        
      function symbol_no(symbol)
      implicit none
      integer :: symbol, symbol_no
      symbol_no = 1
      do while (symbol /= elements_numbers(symbol_no))
            symbol_no = symbol_no + 1
      enddo
      end function symbol_no
      
      function no_nodes_of_prev_atoms(index)
      implicit none
      integer :: index, atom, atom_symbol_no, atom_symbol
      integer :: no_nodes_of_prev_atoms
      
      no_nodes_of_prev_atoms = 0
      do atom = 1, index - 1
        atom_symbol = numbers(atom)
        atom_symbol_no = symbol_no(atom_symbol)
        no_nodes_of_prev_atoms = no_nodes_of_prev_atoms
     *        + no_all_nodes_of_elements(atom_symbol_no)
      enddo
      end function no_nodes_of_prev_atoms
      
      function no_nodes_of_prev_layers(layer_no, layer)
      implicit none
      integer :: layer_no, layer, i
      integer :: no_nodes_of_prev_layers
      
      no_nodes_of_prev_layers = 0
      do i = layer_no + 1, layer_no + layer - 1
        no_nodes_of_prev_layers = no_nodes_of_prev_layers
     *        + no_nodes_of_elements(i)
      enddo
      end function no_nodes_of_prev_layers
      
      function no_prev_layers_of_all_elements(symbol_no)
      implicit none
      integer :: symbol_no, no_prev_layers_of_all_elements
      
      no_prev_layers_of_all_elements = 0
      do i = 1, symbol_no - 1
       no_prev_layers_of_all_elements = no_prev_layers_of_all_elements +
     *                            no_layers_of_elements(i)
      enddo
      end function  no_prev_layers_of_all_elements
      
      function  unravel_matrix(matrix, no_rows, no_cols)
      implicit none
      integer :: no_rows, no_cols
      double precision, dimension (no_rows, no_cols) :: matrix
      double precision, dimension (no_rows*no_cols) :: unravel_matrix
      
       do i = 1, no_rows
        do j = 1, no_cols
            count = (i-1)*no_cols+j
            unravel_matrix(count) = matrix(i,j)
        enddo
      enddo
      end function unravel_matrix
       
      end subroutine der_parameters
          
       subroutine calculate_der_g2(n_indices, numbers, rs, g_number,
     *                        g_eta, cutoff, aa, home, mm, ii, n, ridge)
               
              implicit none
              integer, dimension(n) :: n_indices, numbers
              integer, dimension(1) :: g_number
              double precision, dimension(n, 3) :: rs
              double precision, dimension(3) :: home, Rj
              integer :: n, mm, ii, aa
              double precision ::  g_eta, cutoff
              double precision :: ridge
cf2py         intent(in) :: n_indices, numbers, rs, g_number
cf2py         intent(in) :: g_eta, cutoff, aa, home, mm, ii
cf2py         intent(hide) :: n
cf2py         intent(out) :: ridge
              integer :: j, match, xyz
              double precision, dimension(3) :: Raj_
              double precision :: Raj, term, term1, term2

              ridge = 0.0d0
              do j = 1, n
                  match = compare(numbers(j), g_number(1))
                  if (match == 1) then
                    do xyz = 1, 3
                      Rj(xyz) = rs(j, xyz)
                      Raj_(xyz) = home(xyz) - Rj(xyz)
                    end do
                    Raj = sqrt(dot_product(Raj_, Raj_))
                    term1 = - 2.0d0 * g_eta * Raj * 
     *             cutoff_fxn(Raj, cutoff) + der_cutoff_fxn(Raj, cutoff)
                    term2 = 
     *               der_position(aa, n_indices(j), home, Rj, mm, ii)
                    term = exp(- g_eta * (Raj**2.0d0)) * term1 * term2
                    ridge = ridge + term
                  end if
                end do

      CONTAINS

      function compare(try, val) result(match)
C     Returns 1 if try is the same set as val, 0 if
C     not.
              implicit none
              integer, intent(in) :: try, val
              integer :: match
              if (try == val) then
                      match = 1
              else
                      match = 0
              end if
      end function compare
      
      function cutoff_fxn(r, cutoff)
              double precision :: r, cutoff, cutoff_fxn, pi
              if (r > cutoff) then
                      cutoff_fxn = 0.0d0
              else
                      pi = 4.0d0 * datan(1.0d0)
                      cutoff_fxn = 0.5d0 * (cos(pi*r/cutoff) + 1.0d0)
              end if
      end function
      
      function der_cutoff_fxn(r, cutoff)
              double precision :: r, cutoff, der_cutoff_fxn, pi
              if (r > cutoff) then
                      der_cutoff_fxn = 0.0d0
              else
                      pi = 4.0d0 * datan(1.0d0)
                      der_cutoff_fxn = -0.5d0 * pi * sin(pi*r/cutoff)
     *                      / cutoff
              end if
      end function
      
      function der_position(mm, nn, Rm, Rn, ll, ii)
              integer mm, nn, ll, ii
              double precision, dimension(3) :: Rm, Rn, Rmn_
              double precision :: der_position, Rmn
              do xyz = 1, 3
                      Rmn_(xyz) = Rm(xyz) - Rn(xyz)
              end do
              Rmn = sqrt(dot_product(Rmn_, Rmn_))
              if (ll == mm) then
                      der_position = (Rm(ii + 1) - Rn(ii + 1)) / Rmn
              else if (ll == nn) then
                      der_position = - (Rm(ii + 1) - Rn(ii + 1)) / Rmn
              else
                      der_position = 0.0d0
              end if
      end function
      
      end subroutine calculate_der_g2
 
      subroutine calculate_der_g4(n_indices, numbers, rs, g_numbers,
     *                        g_gamma, g_zeta, 
     *                        g_eta, cutoff, aa, home, mm, ii, n, ridge)
               
              implicit none
              integer, dimension(n) :: n_indices, numbers
              integer, dimension(2) :: g_numbers
              double precision, dimension(n, 3) :: rs
              double precision, dimension(3) :: home, Rj, Rk
              integer :: n, aa, mm, ii
              double precision :: g_gamma, g_zeta, g_eta, cutoff
              double precision :: ridge
cf2py         intent(in) :: numbers, rs, g_numbers, g_gamma, g_zeta
cf2py         intent(in) :: g_eta, cutoff, home, n_indices , aa, mm, ii
cf2py         intent(hide) :: n
cf2py         intent(out) :: ridge
              integer :: j, k, match, xyz
              double precision, dimension(3) :: Raj_, Rak_, Rjk_
             double precision :: Raj, Rak, Rjk, costheta, c1, c2, c3, c4
              double precision :: term1, term2, term3, term4, term5
              double precision :: term6, term7, term8, term9, term10
              double precision :: term11, term

              ridge = 0.0d0
              do j = 1, n
                do k = (j + 1), n
                  match = compare(numbers(j), numbers(k), g_numbers(1),
     *                            g_numbers(2))
                  if (match == 1) then
                    do xyz = 1, 3
                      Rj(xyz) = rs(j, xyz)
                      Rk(xyz) = rs(k, xyz)
                      Raj_(xyz) = Rj(xyz) - home(xyz)
                      Rak_(xyz) = Rk(xyz) - home(xyz)
                      Rjk_(xyz) = Rj(xyz) - Rk(xyz)
                    end do
                    Raj = sqrt(dot_product(Raj_, Raj_))
                    Rak = sqrt(dot_product(Rak_, Rak_))
                    Rjk = sqrt(dot_product(Rjk_, Rjk_))
                    costheta = dot_product(Raj_, Rak_) / Raj / Rak
                    c1 = (1.0d0 + g_gamma * costheta)
                    c2 = cutoff_fxn(Raj, cutoff)
                    c3 = cutoff_fxn(Rak, cutoff)
                    c4 = cutoff_fxn(Rjk, cutoff)
                    if (g_zeta == 1) then
                        term1 = exp(-g_eta*(Raj**2 + Rak**2 + Rjk**2))
                    else
                        term1 = (c1**(g_zeta - 1.0d0))
     *                        * exp(-g_eta*(Raj**2 + Rak**2 + Rjk**2))
                    end if
                    term2 = c2 * c3 * c4
                    term3 = der_cos_theta(aa, n_indices(j), 
     *                n_indices(k), home, Rj, Rk, mm, ii)
                    term4 = g_gamma * g_zeta * term3
                    term5 =
     *               der_position(aa, n_indices(j), home, Rj, mm, ii)
                    term4 = term4 - 2.0d0 * c1 * g_eta * Raj * term5
                    term6 = 
     *               der_position(aa, n_indices(k), home, Rk, mm, ii)
                    term4 = term4 - 2.0d0 * c1 * g_eta * Rak * term6
                    term7 =  der_position(n_indices(j), n_indices(k),
     *                                                 Rj, Rk, mm, ii)
                    term4 = term4 - 2.0d0 * c1 * g_eta * Rjk * term7
                    term2 = term2 * term4
                    term8 = 
     *               c1 * der_cutoff_fxn(Raj, cutoff) * c3 * c4 * term5
                    term9 = 
     *              c1 * c2 * der_cutoff_fxn(Rak, cutoff) * c4 * term6
                    term10 = 
     *              c1 * c2 * c3 * der_cutoff_fxn(Rjk, cutoff) * term7
                    term11 = term2 + term8 + term9 + term10
                    term = term1 * term11
                    ridge = ridge + term
                  end if
                end do
              end do
              ridge = ridge * (2.0d0**(1.0d0 - g_zeta))

      CONTAINS

      function compare(try1, try2, val1, val2) result(match)
C     Returns 1 if (try1, try2) is the same set as (val1, val2), 0 if
C     not.
              implicit none
              integer, intent(in) :: try1, try2, val1, val2
              integer :: match
              integer :: ntry1, ntry2, nval1, nval2
              ! First sort to avoid endless logical loops.
              if (try1 < try2) then
                      ntry1 = try1
                      ntry2 = try2
              else
                      ntry1 = try2
                      ntry2 = try1
              end if
              if (val1 < val2) then
                      nval1 = val1
                      nval2 = val2
              else
                      nval1 = val2
                      nval2 = val1
              end if
              if (ntry1 == nval1 .AND. ntry2 == nval2) then
                      match = 1
              else
                      match = 0
              end if
      end function compare

      function cutoff_fxn(r, cutoff)
              double precision :: r, cutoff, cutoff_fxn, pi
              if (r > cutoff) then
                      cutoff_fxn = 0.0d0
              else
                      pi = 4.0d0 * datan(1.0d0)
                      cutoff_fxn = 0.5d0 * (cos(pi*r/cutoff) + 1.0d0)
              end if
      end function
      
      function der_cutoff_fxn(r, cutoff)
              double precision :: r, cutoff, der_cutoff_fxn, pi
              if (r > cutoff) then
                      der_cutoff_fxn = 0.0d0
              else
                      pi = 4.0d0 * datan(1.0d0)
                      der_cutoff_fxn = -0.5d0 * pi * sin(pi*r/cutoff)
     *                      / cutoff
              end if
      end function
      
      function der_position(mm, nn, Rm, Rn, ll, ii)
              integer :: mm, nn, ll, ii
              double precision, dimension(3) :: Rm, Rn, Rmn_
              double precision :: der_position, Rmn
              do xyz = 1, 3
                      Rmn_(xyz) = Rm(xyz) - Rn(xyz)
              end do
              Rmn = sqrt(dot_product(Rmn_, Rmn_))
              if (ll == mm) then
                      der_position = (Rm(ii + 1) - Rn(ii + 1)) / Rmn
              else if (ll == nn) then
                      der_position = - (Rm(ii + 1) - Rn(ii + 1)) / Rmn
              else
                      der_position = 0.0d0
              end if
      end function
      
      function der_cos_theta(aa, jj, kk, home, Rj, Rk, mm, ii)
      implicit none
      integer :: aa, jj, kk, mm, ii
      double precision :: der_cos_theta
      double precision, dimension(3) :: home, Rj, Rk
      
      do xyz = 1, 3
            Raj_(xyz) = home(xyz) - Rj(xyz)
            Rak_(xyz) = home(xyz) - Rk(xyz)
      end do
      Raj = sqrt(dot_product(Raj_, Raj_))
      Rak = sqrt(dot_product(Rak_, Rak_))
      der_cos_theta = 1.0d0 / (Raj * Rak) *
     *        dot_product(der_position_vector(aa, jj, mm, ii), Rak_)
      der_cos_theta =  der_cos_theta + 1.0d0 / (Raj * Rak) *
     *        dot_product(der_position_vector(aa, kk, mm, ii), Raj_)
      der_cos_theta =  der_cos_theta - 1.0d0 / (Raj * Raj * Rak) *
     *        dot_product(Raj_, Rak_) * 
     *        der_position(aa, jj, home, Rj, mm, ii)
      der_cos_theta =  der_cos_theta - 1.0d0 / (Raj * Rak * Rak) *
     *        dot_product(Raj_, Rak_) * 
     *        der_position(aa, kk, home, Rk, mm, ii)
     
      end function
       
      function der_position_vector(aa, bb, mm, ii)
      implicit none
      integer :: aa, bb, mm, ii
      integer, dimension(3) :: der_position_vector
      
      der_position_vector(1) = (Kronecker_delta(mm, aa)
     *     - Kronecker_delta(mm, bb)) * Kronecker_delta(0, ii)
      der_position_vector(2) = (Kronecker_delta(mm, aa)
     *     - Kronecker_delta(mm, bb)) * Kronecker_delta(1, ii)
      der_position_vector(3) = (Kronecker_delta(mm, aa)
     *     - Kronecker_delta(mm, bb)) * Kronecker_delta(2, ii)
    
      end function
       
      function Kronecker_delta(ii, jj)
      implicit none
      integer :: ii, jj
      double precision :: Kronecker_delta
      
      if (ii == jj) then
        Kronecker_delta = 1.0d0
      else
        Kronecker_delta = 0.0d0
      end if
    
      end function

      end subroutine calculate_der_g4
