#!/usr/bin/env python
"""This module contains utilities for use with various aspects of the
neural network calculators."""

import numpy as np

from ase import io


def randomize_images(images, fraction=0.8):
    """Randomly assigns 'fraction' of the images to a training set and
    (1 - 'fraction') to a test set. Returns two lists of ASE images.
    'images' can either be a list of ASE images or a path to an ASE
    trajectory.
    """
    file_opened = False
    if type(images) == str:
        images = io.PickleTrajectory(images, 'r')
        file_opened = True
    trainingsize = int(fraction * len(images))
    testsize = len(images) - trainingsize
    testindices = []
    while len(testindices) < testsize:
        next = np.random.randint(len(images))
        if next not in testindices:
            testindices.append(next)
    testindices.sort()
    trainindices = [index for index in range(len(images)) if index not in
                    testindices]
    train_images = [images[index] for index in trainindices]
    test_images = [images[index] for index in testindices]
    if file_opened:
        images.close()
    return train_images, test_images


class ExtrapolateError(Exception):
    pass


class UntrainedError(Exception):
    pass
